# -*- coding: utf-8 -*-

"""Resources for y0."""

import os

HERE = os.path.abspath(os.path.dirname(__file__))
ASIA_PATH = os.path.join(HERE, "asia.csv")
VIRAL_PATHOGENESIS_PATH = os.path.join(HERE, "viral_pathogenesis.json")
